/**
 * Hany Char Convertor: conversion table handling routines
 * Copyright 1999,.. by Peter Hanecak <hanecak@megaloman.sk>.
 *
 * Here are functions to load conversion table from CT file.
 *
 * @version 0.1.2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 */

#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include "ct.h"

static char	SEPARATOR_1 = '=';
static char	SEPARATOR_2 = ',';
static char	SEPARATOR_3 = ':';
static char	ASCII_BEGIN = '<';
static char	ASCII_END = '>';
static char	COMMENT_BEGIN = '#';
static char	CONTINUE_LINE = '\\';

static char	ERR_OPEN[] = "can't open file";
static char	ERR_FORMAT[] = "incorrect fotmat";
static char	ERR_FORMAT_TAG_L[] = "incorrect left tag syntax";
static char	ERR_FORMAT_TAG_R[] = "incorrect right tag syntax";
static char	ERR_NUM_FORMAT_TAG_L[] = "incorrect left tag number syntax";
static char	ERR_NUM_FORMAT_TAG_R[] = "incorrect right tag number syntax";
static char	ERR_CT_SMALL[] = "conversion table is too small";
static char	ERR_UNKNOWN_CT[] = "unknown conversion table";

char	*errorMsg = NULL;	// this variable is set when some error occurs
char	*errorParam = NULL;	// detto

t_ct loadct(const char *ctfn, const char *ctn) {
	std::ifstream	*fr;
	char	buffer1[MAXLL] = "";
	char	buffer2[MAXLL] = "";
	char	*line = buffer1;
	char	*lTag, *rTag;
	int	lc, rc;
	char	*temp, *temp2 = line;
	t_ct	ct = NULL;

	// open conversion table file
	#ifdef VERBOSE
	std::cerr << "*** opening " << ctfn << " ..." << std::endl;
	#endif
	//fr = new std::ifstream(ctfn, ios::nocreate | ios::in | ios::binary);
	fr = new std::ifstream(ctfn, std::ios::in | std::ios::binary);
	if (!fr || fr->fail()) {
		errorMsg = ERR_OPEN;
		//errorParam = ctfn;
		return NULL;
	}

	// read convertion table
	errorMsg = ERR_UNKNOWN_CT;	// in case c.t. file do not contain c.t. named "<ctn>"
	//errotParam = ctn;
	while (fr->getline(buffer1, sizeof(buffer1)) != NULL) {
		#ifdef VERBOSE
		std::cerr << "*** buffer1 " << buffer1;
		#endif
		// strip out comments
		temp = index(buffer1, COMMENT_BEGIN);
		if (temp != NULL)
			temp[0] = '\0';
		#ifdef VERBOSE
		std::cerr << " (after stripping comments: " << buffer1 << ")" << std::endl;
		#endif
		temp = index(buffer1, CONTINUE_LINE);
		if (temp != NULL) {
			temp[0] = '\0';
			if (line == buffer2) {
				buffer2[0] = '\0';
				line = buffer1;
			}
			strncat(buffer2, buffer1, sizeof(buffer2));
			continue;
		}
		else {
			if (buffer2[0] != '\0') {
				strncat(buffer2, buffer1, sizeof(buffer2));
				line = buffer2;
			}
		}
		#ifdef VERBOSE
		std::cerr << "*** line " << line << " (after line continuation processing)" << std::endl;
		#endif

		temp = strstr(line, ctn);

		// identify our conversion table line
		if (temp != line)
			continue;
		temp += strlen(ctn);
		if (temp[0] != SEPARATOR_1)
			continue;
		temp += sizeof(SEPARATOR_1);

		// allocate c.t.
		#ifdef VERBOSE
		std::cerr << "*** allocating c.t." << std::endl;
		#endif
		ct = (char *)calloc(ctSize, sizeof(char));
		for (int n1 = 0; n1 < ctSize; n1++)
			ct[n1] = (char)n1;

		// process conversion table line
		#ifdef VERBOSE
		std::cerr << "*** processing c.t.: ";
		#endif
		while (temp2 != NULL) {
			temp2 = index(temp, SEPARATOR_2);	// next tag (or NULL)
			if (temp2 != NULL) {
				temp2[0] = '\0';
				temp2 += sizeof(SEPARATOR_2);
			}

			lTag = temp;
			rTag = index(temp, SEPARATOR_3);
			if (rTag == NULL) {
				errorMsg = ERR_FORMAT;
				errorParam = line;
				return NULL;
			}
			rTag[0] = '\0';
			rTag += sizeof(SEPARATOR_3);

			// left tag
			if (lTag[0] == ASCII_BEGIN) {
				// extended format (i.e. "\<<ASCII>\>")
				if (lTag[strlen(lTag) - 1] != ASCII_END) {
					errorMsg = ERR_FORMAT_TAG_L;
					errorParam = line;
					return NULL;
				}
				lTag[strlen(lTag) - 1] = '\0';
				lTag += sizeof(char);
				lc = strtol(lTag, &temp, 0);
				if (temp[0] != '\0') {
					errorMsg = ERR_NUM_FORMAT_TAG_L;
					errorParam = line;
					return NULL;
				}
			}
			else {
				// simple format (i.e. "<char>")
				lc = lTag[0];
			}

			// right tag
			if (rTag[0] == ASCII_BEGIN) {
				// extended format (i.e. "\<<ASCII>\>")
				if (rTag[strlen(rTag) - 1] != ASCII_END) {
					errorMsg = ERR_FORMAT_TAG_R;
					errorParam = line;
					return NULL;
				}
				rTag[strlen(rTag) - 1] = '\0';
				rTag += sizeof(char);
				rc = strtol(rTag, &temp, 0);
				if (temp[0] != '\0') {
					errorMsg = ERR_NUM_FORMAT_TAG_R;
					errorParam = line;
					return NULL;
				}
			}
			else {
				// simple format (i.e. "<char>")
				rc = rTag[0];
			}

			#ifdef VERBOSE
			std::cerr << lc << ":" << rc << ", ";
			#endif

			// make entry
			if (lc < ctSize)
				ct[lc] = rc;
			else {
				errorMsg = ERR_CT_SMALL;
				errorParam = line;
				return NULL;
			}

			temp = temp2;
		}
		#ifdef VERBOSE
		std::cerr << std::endl;
		#endif

		break;	// first matched c.t. is enought :)
	}

	delete fr;
	return ct;
}
